/*  㠫쭮   C++.
   ࣥ ਭ.
*/

# include <fstream.h>
# include "vp.h"

const int MaxPC = 30000;

void TVirtualProcessor::Init()
{ Code            = 0;
  CodeSize        = 0;
  PC              = 0;
  SP              = 0;
  FromFile        = false;
  Active          = false;
  SubCommand      = 0;
  BasePriority    = 0;
  CurrentPriority = 0;
}

Byte TVirtualProcessor::GetCommand() throw(EVP_BadCode, EVP_BadPC)
{ Byte c;
  if (!Code)
    throw EVP_BadCode();
  if (PC >= CodeSize)
    throw EVP_BadPC();
  c = Code[PC++];
  return c;
}

int TVirtualProcessor::GetInt() throw()
{ int Value;
  Value  = ((int)Code[PC++]     );
  Value |= ((int)Code[PC++] << 8);
  return Value;
}

long TVirtualProcessor::GetLong() throw()
{ long Value;
  Value  = ((int)Code[PC++]      );
  Value |= ((int)Code[PC++] << 8 );
  Value |= ((int)Code[PC++] << 16);
  Value |= ((int)Code[PC++] << 24);
  return Value;
}

void TVirtualProcessor::GoTo(int aPC) throw(EVP_BadPC)
{ if (aPC < 0 || aPC >= CodeSize)
    throw EVP_BadPC();
  PC = aPC;
}

void TVirtualProcessor::Call(int aPC) throw(EVP_StackOverflow, EVP_BadPC)
{ if (SP >= StackSize)
    throw EVP_StackOverflow();
  if (aPC < 0 || aPC >= CodeSize)
    throw EVP_BadPC();
  Stack[SP++] = PC;
  PC = aPC;
}

void TVirtualProcessor::Return() throw(EVP_StackUnderflow)
{ if (SP <= 0)
    throw EVP_StackUnderflow();
  PC = Stack[--SP];
}

void TVirtualProcessor::RemoteCommand()
{ int adr = GetInt();
  int sz  = (int)GetByte();
  TransferCommand(adr, (void*)(Code + PC), sz);
  PC += sz;
}

void TVirtualProcessor::TransferCommand(int, void*, int)
{ throw EVP_BadCommand();
}

TVirtualProcessor::TVirtualProcessor(const Byte* aCode, int aCodeSize)
{ Init();
  if (!aCode || aCodeSize <= 0 || aCodeSize > MaxPC)
    throw EVP_BadCode();
  Code     = const_cast<Byte*>(aCode);
  CodeSize = aCodeSize;
  Start();
}

TVirtualProcessor::TVirtualProcessor(const char* aFile)
{ streampos len;
  ifstream  File;

  Init();
  File.open(aFile, ios::nocreate | ios::binary);
  if (!File)
    throw EVP_FileNotOpen();
  File.seekg(0, ios::end);
  len = File.tellg();
  if (len == 0 || len > MaxPC)
    throw EVP_BadFile();
  CodeSize = (int)len;
  File.seekg(0, ios::beg);
  Code = new Byte[CodeSize];
  File.read(Code, CodeSize);
  if (File.bad())
    throw EVP_FileError();
  File.close();
  if (File.bad())
    throw EVP_FileError();
  Start();
}

TVirtualProcessor::~TVirtualProcessor()
{ if (FromFile && Code)
    delete[] Code;
}

bool TVirtualProcessor::RemoteStep()
{ PC = 0;
  return Step();
}

void TVirtualProcessor::Execute()
{ while (Step())
    ;
}
